import uuid

from typing import TYPE_CHECKING

from sqlalchemy import Boolean, ForeignKey, Integer, Text, Unicode, Uuid
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..db.base_class import reg


if TYPE_CHECKING:
    from .account import Account
    from .product_group import ProductGroup
    from .stock_keeping_unit import StockKeepingUnit


@reg.mapped_as_dataclass(unsafe_hash=True)
class Product:
    __tablename__ = "products"

    id: Mapped[uuid.UUID] = mapped_column(Uuid, primary_key=True, insert_default=uuid.uuid4)
    code: Mapped[int] = mapped_column(Integer, unique=True, nullable=False)
    name: Mapped[str] = mapped_column(Unicode, nullable=False, unique=True)
    description: Mapped[str | None] = mapped_column(Text, nullable=True)
    fraction_units: Mapped[str] = mapped_column(Unicode, nullable=False)
    product_group_id: Mapped[uuid.UUID] = mapped_column(
        Uuid,
        ForeignKey("product_groups.id"),
        nullable=False,
    )
    account_id: Mapped[uuid.UUID] = mapped_column(Uuid, ForeignKey("accounts.id"), nullable=False)
    is_active: Mapped[bool] = mapped_column(Boolean, nullable=False)
    is_fixture: Mapped[bool] = mapped_column(Boolean, nullable=False)
    is_purchased: Mapped[bool] = mapped_column(Boolean, nullable=False)
    is_sold: Mapped[bool] = mapped_column(Boolean, nullable=False)

    skus: Mapped[list["StockKeepingUnit"]] = relationship("StockKeepingUnit", back_populates="product")
    product_group: Mapped["ProductGroup"] = relationship("ProductGroup", back_populates="products")
    account: Mapped["Account"] = relationship("Account", back_populates="products")

    def __init__(
        self,
        name: str,
        fraction_units: str,
        product_group_id: uuid.UUID,
        account_id: uuid.UUID,
        is_active: bool,
        is_purchased: bool,
        is_sold: bool,
        code: int | None = None,
        id_: uuid.UUID | None = None,
        is_fixture: bool | None = False,
    ) -> None:
        if code is not None:
            self.code = code
        self.name = name
        self.fraction_units = fraction_units
        self.product_group_id = product_group_id
        self.account_id = account_id
        self.is_active = is_active
        self.is_purchased = is_purchased
        self.is_sold = is_sold
        if id_ is not None:
            self.id = id_
        if is_fixture is not None:
            self.is_fixture = is_fixture

    @classmethod
    def suspense(cls) -> uuid.UUID:
        return uuid.UUID("aa79a643-9ddc-4790-ac7f-a41f9efb4c15")
