from datetime import date, datetime

import brewman.schemas.purchase_entries as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import select
from sqlalchemy.orm import Session, contains_eager

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.batch import Batch
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.PurchaseEntries)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["purchase-entries"]),
) -> schemas.PurchaseEntries:
    return schemas.PurchaseEntries(
        start_date=get_start_date(request.session),
        finish_date=get_finish_date(request.session),
        body=[],
    )


@router.get("/{start}/{finish}", response_model=schemas.PurchaseEntries)
def report_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["purchase-entries"]),
) -> schemas.PurchaseEntries:
    start_date = datetime.strptime(start, "%d-%b-%Y")
    finish_date = datetime.strptime(finish, "%d-%b-%Y")
    with SessionFuture() as db:
        body = build_report(start_date, finish_date, db)
        set_period(start, finish, request.session)
        return schemas.PurchaseEntries(start_date=start, finish_date=finish, body=body)


def build_report(start_date: date, finish_date: date, db: Session) -> list[schemas.PurchaseEntriesItem]:
    body = []
    query: list[Voucher] = list(
        db.execute(
            select(Voucher)
            .join(Voucher.journals)
            .join(Journal.account)
            .join(Voucher.inventories)
            .join(Inventory.batch)
            .join(Batch.sku)
            .join(StockKeepingUnit.product)
            .options(
                contains_eager(Voucher.journals).contains_eager(Journal.account),
                contains_eager(Voucher.inventories)
                .contains_eager(Inventory.batch)
                .contains_eager(Batch.sku)
                .contains_eager(StockKeepingUnit.product),
            )
            .where(
                Voucher.date_ >= start_date,
                Voucher.date_ <= finish_date,
                Voucher.voucher_type == VoucherType.PURCHASE,
            )
            .order_by(Voucher.date_)
        )
        .unique()
        .scalars()
        .all()
    )

    for voucher in query:
        journal = next(j for j in voucher.journals if j.debit == -1)
        for item in voucher.inventories:
            row = schemas.PurchaseEntriesItem(
                date_=voucher.date_,
                supplier=journal.account.name,
                url=[
                    "/",
                    voucher.voucher_type.name.replace("_", "-").lower(),
                    str(voucher.id),
                ],
                product=item.batch.sku.product.name,
                quantity=item.quantity,
                rate=item.rate,
                tax=item.tax,
                discount=item.discount,
                amount=item.amount,
            )
            body.append(row)
    return body
