import uuid

from datetime import date, datetime
from decimal import Decimal

from pydantic import (
    ConfigDict,
    FieldSerializationInfo,
    field_serializer,
    field_validator,
)
from pydantic.main import BaseModel

from . import to_camel
from .account import AccountLink


class ReconcileItem(BaseModel):
    id_: uuid.UUID | None = None
    date_: date
    name: str
    url: list[str]
    type_: str
    narration: str
    debit: Decimal
    credit: Decimal
    posted: bool
    is_reconciled: bool
    reconcile_date: date | None = None
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("date_")
    def serialize_date(self, value: date, info: FieldSerializationInfo) -> str:
        return value.strftime("%d-%b-%Y")

    @field_validator("reconcile_date", mode="before")
    @classmethod
    def parse_reconcile_date(cls, value: None | date | str) -> date | None:
        if value is None:
            return None
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("reconcile_date")
    def serialize_reconcile_date(self, value: date | None, info: FieldSerializationInfo) -> str | None:
        return None if value is None else value.strftime("%d-%b-%Y")


class Reconcile(BaseModel):
    start_date: date | str
    finish_date: date | str
    account: AccountLink | None = None
    body: list[ReconcileItem]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("start_date", mode="before")
    @classmethod
    def parse_start_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("start_date")
    def serialize_start_date(self, value: date, info: FieldSerializationInfo) -> str:
        return value.strftime("%d-%b-%Y")

    @field_validator("finish_date", mode="before")
    @classmethod
    def parse_finish_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("finish_date")
    def serialize_finish_date(self, value: date, info: FieldSerializationInfo) -> str:
        return value.strftime("%d-%b-%Y")
