import json
import uuid

from datetime import date, datetime
from decimal import Decimal

from fastapi import Form
from pydantic import (
    BaseModel,
    ConfigDict,
    FieldSerializationInfo,
    field_serializer,
    field_validator,
)

from . import to_camel
from .account import AccountLink
from .batch import Batch  # noqa: F401
from .cost_centre import CostCentreLink
from .employee import EmployeeLink  # noqa: F401
from .employee_benefit import EmployeeBenefit
from .image_upload import ImageUpload
from .incentive import Incentive
from .inventory import Inventory
from .journal import Journal
from .product import ProductLink  # noqa: F401
from .user_link import UserLink


class VoucherIn(BaseModel):
    date_: date
    narration: str
    is_starred: bool
    type_: str
    files: list[ImageUpload]

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("date_")
    def serialize_date(self, value: date, info: FieldSerializationInfo) -> str:
        return value.strftime("%d-%b-%Y")

    @classmethod
    def load_form(cls, data: str = Form(...)) -> BaseModel:
        json_data = json.loads(data)
        return cls.parse_obj(json_data)


class Voucher(VoucherIn):
    id_: uuid.UUID | None = None
    is_reconciled: bool | None = None
    reconcile_date: date | None = None
    creation_date: datetime | None = None
    last_edit_date: datetime | None = None
    user: UserLink | None = None
    posted: bool | None = None
    poster: str | None = None
    journals: list[Journal]
    inventories: list[Inventory]
    vendor: AccountLink | None = None
    source: CostCentreLink | None = None
    destination: CostCentreLink | None = None
    incentives: list[Incentive]
    incentive: Decimal | None = None
    employee_benefits: list[EmployeeBenefit]
    files: list[ImageUpload]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("reconcile_date", mode="before")
    @classmethod
    def parse_reconcile_date(cls, value: None | date | str) -> date | None:
        if value is None or value == "":
            return None
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("reconcile_date")
    def serialize_reconcile_date(self, value: date | None, info: FieldSerializationInfo) -> str | None:
        return None if value is None else value.strftime("%d-%b-%Y")

    @field_validator("creation_date", mode="before")
    @classmethod
    def parse_creation_date(cls, value: None | datetime | str) -> datetime | None:
        if value is None or value == "":
            return None
        if isinstance(value, datetime):
            return value
        return datetime.strptime(value, "%d-%b-%Y %H:%M")

    @field_serializer("creation_date")
    def serialize_creation_date(self, value: datetime | None, info: FieldSerializationInfo) -> str | None:
        return None if value is None else value.strftime("%d-%b-%Y %H:%M")

    @field_validator("last_edit_date", mode="before")
    @classmethod
    def parse_last_edit_date(cls, value: None | datetime | str) -> datetime | None:
        if value is None or value == "":
            return None
        if isinstance(value, datetime):
            return value
        return datetime.strptime(value, "%d-%b-%Y %H:%M")

    @field_serializer("last_edit_date")
    def serialize_last_edit_date(self, value: datetime, info: FieldSerializationInfo) -> str | None:
        return None if value is None else value.strftime("%d-%b-%Y %H:%M")
