import { HttpClient, HttpParams } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { Observable } from 'rxjs/internal/Observable';
import { catchError } from 'rxjs/operators';

import { ErrorLoggerService } from '../core/error-logger.service';

import { Employee } from './employee';

const url = '/api/employees';
const serviceName = 'EmployeeService';

@Injectable({ providedIn: 'root' })
export class EmployeeService {
  constructor(
    private http: HttpClient,
    private log: ErrorLoggerService,
  ) {}

  get(id: string | null): Observable<Employee> {
    const getUrl: string = id === null ? `${url}` : `${url}/${id}`;
    return this.http
      .get<Employee>(getUrl)
      .pipe(catchError(this.log.handleError(serviceName, `get id=${id}`))) as Observable<Employee>;
  }

  list(): Observable<Employee[]> {
    return this.http
      .get<Employee[]>(`${url}/list`)
      .pipe(catchError(this.log.handleError(serviceName, 'list'))) as Observable<Employee[]>;
  }

  save(employee: Employee): Observable<Employee> {
    return this.http
      .post<Employee>(`${url}`, employee)
      .pipe(catchError(this.log.handleError(serviceName, 'save'))) as Observable<Employee>;
  }

  update(employee: Employee): Observable<Employee> {
    return this.http
      .put<Employee>(`${url}/${employee.id}`, employee)
      .pipe(catchError(this.log.handleError(serviceName, 'update'))) as Observable<Employee>;
  }

  saveOrUpdate(employee: Employee): Observable<Employee> {
    if (!employee.id) {
      return this.save(employee);
    }
    return this.update(employee);
  }

  delete(id: string): Observable<Employee> {
    return this.http
      .delete<Employee>(`${url}/${id}`)
      .pipe(catchError(this.log.handleError(serviceName, 'delete'))) as Observable<Employee>;
  }

  autocomplete(term: string): Observable<Employee[]> {
    const options = { params: new HttpParams().set('q', term) };
    return this.http
      .get<Employee[]>(`${url}/query`, options)
      .pipe(catchError(this.log.handleError(serviceName, 'autocomplete'))) as Observable<
      Employee[]
    >;
  }
}
